"""
pygamelib is a module aimed at helping people to learn Python with a fun and
easy game development framework.
It is developped with young kids and total beginners in mind and tries to provide more
advanced mechanics to fuel the growth of the learning developers.
"""

name = "pygamelib"
